from dataclasses import dataclass, field
from typing import List, Optional

from alitra import Pose

from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.status import MissionStatus
from robot_interface.models.mission.task import TASKS
from robot_interface.utilities.uuid_string_factory import uuid4_string


@dataclass
class Mission:
    tasks: List[TASKS]
    id: str = field(default_factory=uuid4_string, init=True)
    name: str = ""
    start_pose: Optional[Pose] = None
    dock: Optional[bool] = None
    undock: Optional[bool] = None
    status: MissionStatus = MissionStatus.NotStarted
    error_message: Optional[ErrorMessage] = field(default=None, init=False)

    def _set_unique_id(self) -> None:
        self.id: str = uuid4_string()

    def __post_init__(self) -> None:
        if self.id is None:
            self._set_unique_id()



import logging
from importlib import import_module
from logging import Logger
from types import ModuleType
from typing import Dict, List, Tuple, Union

from injector import Injector, Module, multiprovider, provider, singleton

from isar.apis.api import API
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.robot_control.robot_controller import RobotController
from isar.apis.security.authentication import Authenticator
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.settings import settings
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerInterface
from isar.mission_planner.sequential_task_selector import SequentialTaskSelector
from isar.mission_planner.task_selector_interface import TaskSelectorInterface
from isar.models.communication.queues.queues import Queues
from isar.services.service_connections.request_handler import RequestHandler
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.services.utilities.robot_utilities import RobotUtilities
from isar.state_machine.state_machine import StateMachine
from isar.storage.blob_storage import BlobStorage
from isar.storage.local_storage import LocalStorage
from isar.storage.slimm_storage import SlimmStorage
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface, MqttPublisher


class APIModule(Module):
    @provider
    @singleton
    def provide_api(
        self,
        authenticator: Authenticator,
        scheduling_controller: SchedulingController,
        robot_controller: RobotController,
        keyvault: Keyvault,
    ) -> API:
        return API(authenticator, scheduling_controller, robot_controller, keyvault)

    @provider
    @singleton
    def provide_scheduling_controller(
        self,
        scheduling_utilities: SchedulingUtilities,
    ) -> SchedulingController:
        return SchedulingController(scheduling_utilities)

    @provider
    @singleton
    def provide_robot_controller(
        self,
        robot_utilities: RobotUtilities,
    ) -> RobotController:
        return RobotController(robot_utilities)


class AuthenticationModule(Module):
    @provider
    @singleton
    def provide_authenticator(self) -> Authenticator:
        return Authenticator()


class RobotModule(Module):
    @provider
    @singleton
    def provide_robot_interface(self) -> RobotInterface:
        robot_package_name: str = settings.ROBOT_PACKAGE
        robot: ModuleType = import_module(robot_package_name)
        return robot.robotinterface.Robot()  # type: ignore


class QueuesModule(Module):
    @provider
    @singleton
    def provide_queues(self) -> Queues:
        return Queues()


class RequestHandlerModule(Module):
    @provider
    @singleton
    def provide_request_handler(self) -> RequestHandler:
        return RequestHandler()


class BlobStorageModule(Module):
    @multiprovider
    @singleton
    def provide_blob_storage(self, keyvault: Keyvault) -> List[StorageInterface]:
        return [BlobStorage(keyvault)]


class LocalStorageModule(Module):
    @multiprovider
    @singleton
    def provide_local_storage(self) -> List[StorageInterface]:
        return [LocalStorage()]


class SlimmStorageModule(Module):
    @multiprovider
    @singleton
    def provide_slimm_storage(
        self, request_handler: RequestHandler
    ) -> List[StorageInterface]:
        return [SlimmStorage(request_handler=request_handler)]


class LocalPlannerModule(Module):
    @provider
    @singleton
    def provide_local_planner(self) -> MissionPlannerInterface:
        return LocalPlanner()


class StateMachineModule(Module):
    @provider
    @singleton
    def provide_state_machine(
        self,
        queues: Queues,
        robot: RobotInterface,
        mqtt_client: MqttClientInterface,
        task_selector: TaskSelectorInterface,
    ) -> StateMachine:
        return StateMachine(
            queues=queues,
            robot=robot,
            mqtt_publisher=mqtt_client,
            task_selector=task_selector,
        )


class UploaderModule(Module):
    @provider
    @singleton
    def provide_uploader(
        self,
        queues: Queues,
        storage_handlers: List[StorageInterface],
        mqtt_client: MqttClientInterface,
    ) -> Uploader:
        return Uploader(
            queues=queues,
            storage_handlers=storage_handlers,
            mqtt_publisher=mqtt_client,
        )


class SchedulingUtilitiesModule(Module):
    @provider
    @singleton
    def provide_scheduling_utilities(
        self, queues: Queues, mission_planner: MissionPlannerInterface
    ) -> SchedulingUtilities:
        return SchedulingUtilities(queues, mission_planner)


class RobotUtilitiesModule(Module):
    @provider
    @singleton
    def provide_robot_utilities(self, robot: RobotInterface) -> RobotUtilities:
        return RobotUtilities(robot)


class ServiceModule(Module):
    @provider
    @singleton
    def provide_keyvault(self) -> Keyvault:
        return Keyvault(keyvault_name=settings.KEYVAULT_NAME)


class MqttModule(Module):
    @provider
    @singleton
    def provide_mqtt_client(self, queues: Queues) -> MqttClientInterface:
        if settings.MQTT_ENABLED:
            return MqttPublisher(mqtt_queue=queues.mqtt_queue)
        return None


class SequentialTaskSelectorModule(Module):
    @provider
    @singleton
    def provide_task_selector(self) -> TaskSelectorInterface:
        return SequentialTaskSelector()


modules: Dict[str, Tuple[Module, Union[str, bool]]] = {
    "api": (APIModule, "required"),
    "authentication": (AuthenticationModule, "required"),
    "queues": (QueuesModule, "required"),
    "request_handler": (RequestHandlerModule, "required"),
    "robot_package": (RobotModule, settings.ROBOT_PACKAGE),
    "isar_id": (RobotModule, settings.ISAR_ID),
    "robot_name": (RobotModule, settings.ROBOT_NAME),
    "mission_planner": (LocalPlannerModule, settings.MISSION_PLANNER),
    "task_selector": (
        {"sequential": SequentialTaskSelectorModule}[settings.TASK_SELECTOR],
        settings.TASK_SELECTOR,
    ),
    "service": (ServiceModule, "required"),
    "state_machine": (StateMachineModule, "required"),
    "storage_local": (LocalStorageModule, settings.STORAGE_LOCAL_ENABLED),
    "storage_blob": (BlobStorageModule, settings.STORAGE_BLOB_ENABLED),
    "storage_slimm": (SlimmStorageModule, settings.STORAGE_SLIMM_ENABLED),
    "mqtt": (MqttModule, "required"),
    "utilities": (SchedulingUtilitiesModule, "required"),
    "robot_utilities": (RobotUtilitiesModule, "required"),
}


def get_injector() -> Injector:
    injector_modules: List[Module] = []
    module_overview: str = ""

    for category, (module, config_option) in modules.items():
        if config_option:
            injector_modules.append(module)
        module_overview += (
            f"\n    {category:<15} : {config_option:<20} ({module.__name__})"
        )

    logger: Logger = logging.getLogger("modules")
    logger.info(f"Loaded the following module configurations:{module_overview}")

    return Injector(injector_modules)



# This file is added for backwards-compability for versions 1.19.2 and older.
# It can be removed in the next minor bump (1.20.0)

from isar.script import start

if __name__ == "__main__":
    start()



import logging
import time
from typing import TYPE_CHECKING, Callable, Optional

from transitions import State

from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotActionException,
    RobotException,
)

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Stop(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="stop", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.stop_thread: Optional[ThreadedRequest] = None
        self._count_number_retries: int = 0

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        if self.stop_thread:
            self.stop_thread.wait_for_thread()
        self.stop_thread = None
        self._count_number_retries = 0

    def _run(self) -> None:
        transition: Callable
        while True:
            if not self.stop_thread:
                self.stop_thread = ThreadedRequest(self.state_machine.robot.stop)
                self.stop_thread.start_thread(name="State Machine Stop Robot")

            if self.state_machine.should_stop_mission():
                self.state_machine.stopped = True

            try:
                self.stop_thread.get_output()
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue

            except (RobotActionException, RobotException) as e:
                if self.handle_stop_fail(
                    retry_limit=self.state_machine.stop_robot_attempts_limit,
                    error_message=ErrorMessage(
                        error_reason=e.error_reason,
                        error_description=e.error_description,
                    ),
                ):
                    transition = self.state_machine.mission_stopped  # type: ignore
                    break

                self.logger.warning(
                    f"\nFailed to stop robot because: {e.error_description}"
                    f"\nAttempting to stop the robot again"
                )

                self.stop_thread = None
                continue
            if self.state_machine.stopped:
                transition = self.state_machine.mission_stopped  # type: ignore
            else:
                transition = self.state_machine.mission_paused  # type: ignore
            break

        transition()

    def handle_stop_fail(self, retry_limit: int, error_message: ErrorMessage) -> bool:
        self._count_number_retries += 1
        if self._count_number_retries > retry_limit:
            self.state_machine.current_task.error_message = error_message

            self.logger.error(
                f"\nFailed to stop the robot after {retry_limit} attempts because: "
                f"{error_message.error_description}"
                f"\nBe aware that the robot may still be moving even though a stop has "
                "been attempted"
            )

            return True
        time.sleep(self.state_machine.sleep_time)
        return False



import logging
import logging.config
from importlib.resources import as_file, files

import yaml
from opencensus.ext.azure.log_exporter import AzureLogHandler
from uvicorn.logging import ColourizedFormatter

from isar.config.configuration_error import ConfigurationError
from isar.config.keyvault.keyvault_error import KeyvaultError
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.settings import settings


def setup_loggers(keyvault: Keyvault) -> None:
    log_levels: dict = settings.LOG_LEVELS
    source = files("isar").joinpath("config").joinpath("logging.conf")
    with as_file(source) as f:
        log_config = yaml.safe_load(open(f))

    logging.config.dictConfig(log_config)

    handlers = []
    if settings.LOG_HANDLER_LOCAL_ENABLED:
        handlers.append(configure_console_handler(log_config=log_config))
    if settings.LOG_HANDLER_APPLICATION_INSIGHTS_ENABLED:
        handlers.append(
            configure_azure_handler(log_config=log_config, keyvault=keyvault)
        )

    for log_handler in handlers:
        for loggers in log_config["loggers"].keys():
            logging.getLogger(loggers).addHandler(log_handler)
            logging.getLogger(loggers).setLevel(log_levels[loggers])
        logging.getLogger().addHandler(log_handler)


def configure_console_handler(log_config: dict) -> logging.Handler:
    handler = logging.StreamHandler()
    handler.setLevel(log_config["root"]["level"])
    handler.setFormatter(
        ColourizedFormatter(
            log_config["formatters"]["colourized"]["format"],
            style="{",
            use_colors=True,
        )
    )
    return handler


def configure_azure_handler(log_config: dict, keyvault: Keyvault) -> logging.Handler:
    connection_string: str
    try:
        connection_string = keyvault.get_secret(
            "application-insights-connection-string"
        ).value
    except KeyvaultError:
        message: str = (
            f"CRITICAL ERROR: Missing connection string for Application Insights in key vault '{keyvault.name}'."
        )
        print(f"\n{message} \n")
        raise ConfigurationError(message)

    handler = AzureLogHandler(connection_string=connection_string)
    handler.setLevel(log_config["root"]["level"])
    return handler



import logging
import sys
import time
from logging import Logger
from threading import Thread
from typing import Any, List, Tuple

from injector import Injector

import isar
from isar.apis.api import API
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.log import setup_loggers
from isar.config.settings import robot_settings, settings
from isar.models.communication.queues.queues import Queues
from isar.modules import get_injector
from isar.services.service_connections.mqtt.mqtt_client import MqttClient
from isar.services.service_connections.mqtt.robot_heartbeat_publisher import (
    RobotHeartbeatPublisher,
)
from isar.services.service_connections.mqtt.robot_info_publisher import (
    RobotInfoPublisher,
)
from isar.state_machine.state_machine import StateMachine, main
from isar.storage.uploader import Uploader
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.robot_interface import RobotInterface


def print_setting(
    setting: str = "", value: Any = "", fillchar: str = " ", width: int = 48
):
    separator = ": " if value != "" else ""
    text = setting.ljust(22, fillchar) + separator + str(value)
    print("*", text.ljust(width - 4, fillchar), "*")


def print_startup_info():
    print(
        """
         __   ________   ___        ________
        / /  / ______/  /   |      / ____  /
       / /  / /_____   / /| |     / /___/ /
      / /  /_____  /  / __  |    / __  __/
     / /  ______/ /  / /  | |   / /  | |
    /_/  /_______/  /_/   |_|  /_/   |_|

"""
    )

    WIDTH = 48

    def print_setting(setting: str = "", value: Any = "", fillchar: str = " "):
        separator = ": " if value != "" else ""
        text = setting.ljust(22, fillchar) + separator + str(value)
        print("*", text.ljust(WIDTH - 4, fillchar), "*")

    print("Integration and Supervisory control".center(WIDTH, " "))
    print("of Autonomous Robots".center(WIDTH, " "))
    print()
    print(f"Version: {isar.__version__}\n".center(WIDTH, " "))

    print_setting(fillchar="*")
    print_setting("ISAR settings")
    print_setting(fillchar="-")
    print_setting("Robot package", settings.ROBOT_PACKAGE)
    print_setting("Robot name", settings.ROBOT_NAME)
    print_setting("Run mission stepwise", settings.RUN_MISSION_BY_TASK)
    print_setting("Running on port", settings.API_PORT)
    print_setting("Mission planner", settings.MISSION_PLANNER)
    print_setting("Using local storage", settings.STORAGE_LOCAL_ENABLED)
    print_setting("Using blob storage", settings.STORAGE_BLOB_ENABLED)
    print_setting("Using SLIMM storage", settings.STORAGE_SLIMM_ENABLED)
    print_setting("Using async inspection uploading", settings.UPLOAD_INSPECTIONS_ASYNC)
    print_setting("Plant code", settings.PLANT_CODE)
    print_setting("Plant name", settings.PLANT_NAME)
    print_setting("Plant shortname", settings.PLANT_SHORT_NAME)
    print_setting(fillchar="-")
    print_setting("Robot capabilities", robot_settings.CAPABILITIES)
    print_setting(fillchar="*")
    print()


def start():
    injector: Injector = get_injector()

    keyvault_client = injector.get(Keyvault)
    setup_loggers(keyvault=keyvault_client)
    logger: Logger = logging.getLogger("main")

    print_startup_info()

    state_machine: StateMachine = injector.get(StateMachine)
    uploader: Uploader = injector.get(Uploader)
    robot: RobotInterface = injector.get(RobotInterface)
    queues: Queues = injector.get(Queues)

    threads: List[Thread] = []

    state_machine_thread: Thread = Thread(
        target=main, name="ISAR State Machine", args=[state_machine], daemon=True
    )
    threads.append(state_machine_thread)

    uploader_thread: Thread = Thread(
        target=uploader.run, name="ISAR Uploader", daemon=True
    )
    threads.append(uploader_thread)

    if settings.UPLOAD_INSPECTIONS_ASYNC:

        def inspections_callback(inspection: Inspection, mission: Mission):
            message: Tuple[Inspection, Mission] = (
                inspection,
                mission,
            )
            state_machine.queues.upload_queue.put(message)

        robot.register_inspection_callback(inspections_callback)

    if settings.MQTT_ENABLED:
        mqtt_client: MqttClient = MqttClient(mqtt_queue=queues.mqtt_queue)

        mqtt_thread: Thread = Thread(
            target=mqtt_client.run, name="ISAR MQTT Client", daemon=True
        )
        threads.append(mqtt_thread)

        robot_info_publisher: RobotInfoPublisher = RobotInfoPublisher(
            mqtt_queue=queues.mqtt_queue
        )
        robot_info_thread: Thread = Thread(
            target=robot_info_publisher.run,
            name="ISAR Robot Info Publisher",
            daemon=True,
        )
        threads.append(robot_info_thread)

        robot_heartbeat_publisher: RobotHeartbeatPublisher = RobotHeartbeatPublisher(
            mqtt_queue=queues.mqtt_queue
        )

        robot_heartbeat_thread: Thread = Thread(
            target=robot_heartbeat_publisher.run,
            name="ISAR Robot Heartbeat Publisher",
            daemon=True,
        )
        threads.append(robot_heartbeat_thread)

        publishers: List[Thread] = robot.get_telemetry_publishers(
            queue=queues.mqtt_queue,
            robot_name=settings.ROBOT_NAME,
            isar_id=settings.ISAR_ID,
        )

        if publishers:
            threads.extend(publishers)

    api: API = injector.get(API)
    api_thread: Thread = Thread(target=api.run_app, name="ISAR API", daemon=True)
    threads.append(api_thread)

    for thread in threads:
        thread.start()
        logger.info(f"Started thread: {thread.name}")

    while True:
        for thread in threads:
            if not thread.is_alive():
                logger.critical("Thread '%s' failed - ISAR shutting down", thread.name)
                sys.exit(1)
        time.sleep(state_machine.sleep_time)



import time
from enum import Enum
from typing import List, Optional

from alitra import Frame, Orientation, Pose, Position
from pydantic import BaseModel, Field

from isar.apis.models.models import InputPose, InputPosition
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import (
    TASKS,
    DockingProcedure,
    Localize,
    RecordAudio,
    ReturnToHome,
    TakeImage,
    TakeThermalImage,
    TakeThermalVideo,
    TakeVideo,
    ZoomDescription,
)


class InspectionTypes(str, Enum):
    image: str = "Image"
    thermal_image: str = "ThermalImage"
    video: str = "Video"
    thermal_video: str = "ThermalVideo"
    audio: str = "Audio"


class TaskType(str, Enum):
    Inspection: str = "inspection"
    Localization: str = "localization"
    ReturnToHome: str = "return_to_home"
    Dock: str = "dock"


class StartMissionInspectionDefinition(BaseModel):
    type: InspectionTypes = Field(default=InspectionTypes.image)
    inspection_target: InputPosition
    analysis_type: Optional[str] = None
    duration: Optional[float] = None
    metadata: Optional[dict] = None
    id: Optional[str] = None


class StartMissionTaskDefinition(BaseModel):
    type: TaskType = Field(default=TaskType.Inspection)
    pose: InputPose
    inspection: Optional[StartMissionInspectionDefinition] = None
    tag: Optional[str] = None
    id: Optional[str] = None
    zoom: Optional[ZoomDescription] = None


class StartMissionDefinition(BaseModel):
    tasks: List[StartMissionTaskDefinition]
    id: Optional[str] = None
    name: Optional[str] = None
    start_pose: Optional[InputPose] = None
    dock: Optional[bool] = None
    undock: Optional[bool] = None


def to_isar_mission(start_mission_definition: StartMissionDefinition) -> Mission:
    isar_tasks: List[TASKS] = []

    for start_mission_task_definition in start_mission_definition.tasks:
        task: TASKS = create_isar_task(start_mission_task_definition)
        if start_mission_task_definition.id:
            task.id = start_mission_task_definition.id
        isar_tasks.append(task)

    if not isar_tasks:
        raise MissionPlannerError("Mission does not contain any valid tasks")

    check_for_duplicate_ids(isar_tasks)

    isar_mission: Mission = Mission(tasks=isar_tasks)

    isar_mission.dock = start_mission_definition.dock
    isar_mission.undock = start_mission_definition.undock

    if start_mission_definition.name:
        isar_mission.name = start_mission_definition.name
    else:
        isar_mission.name = _build_mission_name()

    if start_mission_definition.id:
        isar_mission.id = start_mission_definition.id

    if start_mission_definition.start_pose:
        input_pose: InputPose = start_mission_definition.start_pose
        input_frame: Frame = Frame(name=input_pose.frame_name)
        input_position: Position = Position(
            input_pose.position.x,
            input_pose.position.y,
            input_pose.position.z,
            input_frame,
        )
        input_orientation: Orientation = Orientation(
            input_pose.orientation.x,
            input_pose.orientation.y,
            input_pose.orientation.z,
            input_pose.orientation.w,
            input_frame,
        )
        isar_mission.start_pose = Pose(
            position=input_position, orientation=input_orientation, frame=input_frame
        )

    return isar_mission


def check_for_duplicate_ids(items: List[TASKS]):
    duplicate_ids = get_duplicate_ids(items=items)
    if len(duplicate_ids) > 0:
        raise MissionPlannerError(
            f"Failed to create as there were duplicate IDs which is not allowed "
            f"({duplicate_ids})"
        )


def create_isar_task(start_mission_task_definition) -> TASKS:

    if start_mission_task_definition.type == TaskType.Inspection:
        return create_inspection_task(start_mission_task_definition)
    elif start_mission_task_definition.type == TaskType.Localization:
        return create_localization_task(start_mission_task_definition)
    elif start_mission_task_definition.type == TaskType.ReturnToHome:
        return create_return_to_home_task(start_mission_task_definition)
    elif start_mission_task_definition.type == TaskType.Dock:
        return create_dock_task()
    else:
        raise MissionPlannerError(
            f"Failed to create task: '{start_mission_task_definition.type}' is not a valid"
        )


def create_inspection_task(
    start_mission_task_definition: StartMissionTaskDefinition,
) -> TASKS:

    if start_mission_task_definition.inspection.type == InspectionTypes.image:
        return TakeImage(
            target=start_mission_task_definition.inspection.inspection_target.to_alitra_position(),
            tag_id=start_mission_task_definition.tag,
            robot_pose=start_mission_task_definition.pose.to_alitra_pose(),
            metadata=start_mission_task_definition.inspection.metadata,
            zoom=start_mission_task_definition.zoom,
        )
    elif start_mission_task_definition.inspection.type == InspectionTypes.video:
        return TakeVideo(
            target=start_mission_task_definition.inspection.inspection_target.to_alitra_position(),
            duration=start_mission_task_definition.inspection.duration,
            tag_id=start_mission_task_definition.tag,
            robot_pose=start_mission_task_definition.pose.to_alitra_pose(),
            metadata=start_mission_task_definition.inspection.metadata,
            zoom=start_mission_task_definition.zoom,
        )

    elif start_mission_task_definition.inspection.type == InspectionTypes.thermal_image:
        return TakeThermalImage(
            target=start_mission_task_definition.inspection.inspection_target.to_alitra_position(),
            tag_id=start_mission_task_definition.tag,
            robot_pose=start_mission_task_definition.pose.to_alitra_pose(),
            metadata=start_mission_task_definition.inspection.metadata,
            zoom=start_mission_task_definition.zoom,
        )

    elif start_mission_task_definition.inspection.type == InspectionTypes.thermal_video:
        return TakeThermalVideo(
            target=start_mission_task_definition.inspection.inspection_target.to_alitra_position(),
            duration=start_mission_task_definition.inspection.duration,
            tag_id=start_mission_task_definition.tag,
            robot_pose=start_mission_task_definition.pose.to_alitra_pose(),
            metadata=start_mission_task_definition.inspection.metadata,
            zoom=start_mission_task_definition.zoom,
        )

    elif start_mission_task_definition.inspection.type == InspectionTypes.audio:
        return RecordAudio(
            target=start_mission_task_definition.inspection.inspection_target.to_alitra_position(),
            duration=start_mission_task_definition.inspection.duration,
            tag_id=start_mission_task_definition.tag,
            robot_pose=start_mission_task_definition.pose.to_alitra_pose(),
            metadata=start_mission_task_definition.inspection.metadata,
            zoom=start_mission_task_definition.zoom,
        )
    else:
        raise ValueError(
            f"Inspection type '{start_mission_task_definition.inspection.type}' not supported"
        )


def create_localization_task(
    start_mission_task_definition: StartMissionTaskDefinition,
) -> Localize:
    return Localize(
        localization_pose=start_mission_task_definition.pose.to_alitra_pose()
    )


def create_return_to_home_task(
    start_mission_task_definition: StartMissionTaskDefinition,
) -> ReturnToHome:
    return ReturnToHome(pose=start_mission_task_definition.pose.to_alitra_pose())


def create_dock_task() -> DockingProcedure:
    return DockingProcedure(behavior="dock")


def get_duplicate_ids(items: List[TASKS]) -> List[str]:
    unique_ids: List[str] = []
    duplicate_ids: List[str] = []
    for item in items:
        id: str = item.id
        if id not in unique_ids:
            unique_ids.append(id)
        else:
            duplicate_ids.append(id)

    return duplicate_ids


def _build_mission_name() -> str:
    return f"{settings.PLANT_SHORT_NAME}{settings.ROBOT_NAME}{int(time.time())}"



from queue import Queue

from isar.config.settings import settings
from isar.models.communication.queues.queue_io import QueueIO
from isar.models.communication.queues.status_queue import StatusQueue


class Queues:
    def __init__(self) -> None:
        self.start_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.stop_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.pause_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.resume_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.single_action: QueueIO = QueueIO(input_size=1, output_size=1)
        self.upload_queue: Queue = Queue(maxsize=10)
        self.state: StatusQueue = StatusQueue()

        if settings.MQTT_ENABLED:
            self.mqtt_queue: Queue = Queue()



from dataclasses import asdict
from typing import Any

import pytest
from alitra import Pose

from isar.services.readers.base_reader import BaseReader
from robot_interface.models.mission.task import ReturnToHome, TakeImage
from tests.mocks.pose import MockPose
from tests.mocks.task import MockTask


class TestBaseReader:
    @pytest.mark.parametrize(
        "dataclass_dict, expected_dataclass",
        [
            (asdict(MockTask.return_home()), ReturnToHome),
            (asdict(MockPose.default_pose()), Pose),
        ],
    )
    def test_dict_to_dataclass(self, dataclass_dict: dict, expected_dataclass: Any):
        content = BaseReader.dict_to_dataclass(dataclass_dict, expected_dataclass)
        assert type(content) is expected_dataclass



import json
from http import HTTPStatus

import jwt
import pytest


def mock_access_token():
    token = jwt.encode(payload={}, key="some_key")

    return token


class TestAuthentication:
    @pytest.mark.parametrize(
        "query_string, token, expected_output, expected_status_code",
        [
            (
                "start-mission?ID=1",
                "Dummy Token",
                {"detail": "Invalid token format"},
                HTTPStatus.UNAUTHORIZED,
            ),
            (
                "start-mission?ID=1",
                mock_access_token(),
                {"detail": "Unable to verify token, no signing keys found"},
                HTTPStatus.UNAUTHORIZED,
            ),
            (
                "stop-mission",
                mock_access_token(),
                {"detail": "Unable to verify token, no signing keys found"},
                HTTPStatus.UNAUTHORIZED,
            ),
            (
                "drive-to?x-value=1&y-value=2&z-value=3&quaternion=0&quaternion=0&quaternion=0&quaternion=1",  # noqa: E501
                mock_access_token(),
                {"detail": "Unable to verify token, no signing keys found"},
                HTTPStatus.UNAUTHORIZED,
            ),
        ],
    )
    def test_authentication(
        self,
        client_auth,
        query_string,
        token,
        expected_output,
        expected_status_code,
    ):
        response = client_auth.post(
            f"schedule/{query_string}",
            headers={"Authorization": "Bearer " + token},
        )
        result_message = json.loads(response.text)

        assert result_message == expected_output
        assert response.status_code == expected_status_code



from alitra import Frame, Orientation, Pose, Position

from isar.services.readers.base_reader import BaseReader
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import (
    TASKS,
    TakeImage,
    TakeThermalImage,
    ReturnToHome,
)

robot_pose_1 = Pose(
    position=Position(x=2, y=2, z=0, frame=Frame(name="asset")),
    orientation=Orientation(
        x=0, y=0, z=0.7071068, w=0.7071068, frame=Frame(name="asset")
    ),
    frame=Frame(name="asset"),
)

robot_pose_2 = Pose(
    position=Position(x=4, y=4, z=0, frame=Frame(name="asset")),
    orientation=Orientation(
        x=0, y=0, z=-0.7071068, w=0.7071068, frame=Frame(name="asset")
    ),
    frame=Frame(name="asset"),
)


task_take_image = TakeImage(
    target=Position(x=2, y=3, z=1, frame=Frame(name="asset")),
    robot_pose=robot_pose_1,
)

task_take_thermal_image = TakeThermalImage(
    target=Position(x=4, y=3, z=1, frame=Frame(name="asset")),
    robot_pose=robot_pose_2,
)

task_return_to_home = ReturnToHome(
    pose=Pose(
        position=Position(x=0, y=0, z=0, frame=Frame(name="asset")),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame(name="asset")),
        frame=Frame(name="asset"),
    ),
)

expected_mission = Mission(
    id="1",
    tasks=[task_take_image, task_take_thermal_image, task_return_to_home],
)

example_mission_dict = {
    "id": "1",
    "tasks": [
        {
            "type": "take_image",
            "target": {"x": 2, "y": 3, "z": 1, "frame": {"name": "asset"}},
            "robot_pose": {
                "position": {
                    "x": 2,
                    "y": 2,
                    "z": 0,
                    "frame": {"name": "asset"},
                },
                "orientation": {
                    "x": 0,
                    "y": 0,
                    "z": 0.7071068,
                    "w": 0.7071068,
                    "frame": {"name": "asset"},
                },
                "frame": {"name": "asset"},
            },
        },
        {
            "type": "take_thermal_image",
            "target": {"x": 4, "y": 3, "z": 1, "frame": {"name": "asset"}},
            "pose": {
                "position": {
                    "x": 4,
                    "y": 4,
                    "z": 0,
                    "frame": {"name": "asset"},
                },
                "orientation": {
                    "x": 0,
                    "y": 0,
                    "z": -0.7071068,
                    "w": 0.7071068,
                    "frame": {"name": "asset"},
                },
                "frame": {"name": "asset"},
            },
        },
        {
            "type": "return_to_home",
            "pose": {
                "position": {
                    "x": 0,
                    "y": 0,
                    "z": 0,
                    "frame": {"name": "asset"},
                },
                "orientation": {
                    "x": 0,
                    "y": 0,
                    "z": 0,
                    "w": 1,
                    "frame": {"name": "asset"},
                },
                "frame": {"name": "asset"},
            },
        },
    ],
}


def test_mission_definition() -> None:
    loaded_mission: Mission = BaseReader.dict_to_dataclass(
        dataclass_dict=example_mission_dict,
        target_dataclass=Mission,
        strict_config=False,
    )

    assert loaded_mission.id == expected_mission.id
    assert loaded_mission.status == expected_mission.status

    assert len(loaded_mission.tasks) == len(expected_mission.tasks)
    for i_task in range(len(loaded_mission.tasks)):
        loaded_task: TASKS = loaded_mission.tasks[i_task]
        expected_task: TASKS = expected_mission.tasks[i_task]

        assert loaded_task.status == expected_task.status
        assert loaded_task.tag_id == expected_task.tag_id



from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Thread
from typing import Callable, List

from robot_interface.models.robots.media import MediaConfig
from robot_interface.models.initialize import InitializeParams
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus
from robot_interface.models.mission.task import InspectionTask, Task


class RobotInterface(metaclass=ABCMeta):
    """Interface to communicate with robots."""

    @abstractmethod
    def initiate_mission(self, mission: Mission) -> None:
        """Send a mission to the robot and initiate execution of the mission

        This function should be used in combination with the mission_status function
        if the robot is designed to run full missions and not in a taskwise
        configuration.

        Parameters
        ----------
        mission: Mission

        Returns
        -------
        None

        Raises
        ------
        RobotInfeasibleMissionException
            If the mission input is infeasible and the mission fails to be scheduled in
            a way that means attempting to schedule again is not necessary
        RobotException
            Will catch all RobotExceptions not previously listed and retry scheduling of
            the mission until the number of allowed retries is exceeded
        NotImplementedError
            If the robot is designed for taskwise mission execution

        """
        raise NotImplementedError

    @abstractmethod
    def initiate_task(self, task: Task) -> None:
        """Send a task to the robot and initiate the execution of the task

        This function should be used in combination with the task_status function
        if the robot is designed to run taskwise missions and not in a full mission
        configuration.

        Parameters
        ----------
        task : Task
            The task that should be initiated on the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotInfeasibleTaskException
            If the task input is infeasible and the task fails to be scheduled in
            a way that means attempting to schedule again is not necessary
        RobotException
            Will catch all RobotExceptions not previously listed and retry scheduling
            of the task until the number of allowed retries is exceeded before the task
            will be marked as failed and the mission cancelled
        NotImplementedError
            If the robot is designed for full mission execution.

        """
        raise NotImplementedError

    @abstractmethod
    def task_status(self, task_id: str) -> TaskStatus:
        """Gets the status of the currently active task on robot.

        This function should be used in combination with the initiate_task function
        if the robot is designed to run taskwise missions and not in a full mission
        configuration.

        Returns
        -------
        TaskStatus
            Status of the execution of current task.

        Raises
        ------
        RobotException
            If the task status could not be retrieved.
        NotImplementedError
            If the robot is designed for full mission execution.

        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stops the execution of the current task and stops the movement of the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotActionException
            If the robot fails to perform the requested action to stop mission execution
            the action to stop will be attempted again until a certain number of retries
        RobotException
            Will catch other RobotExceptions and retry to stop the mission

        """
        raise NotImplementedError

    @abstractmethod
    def pause(self) -> None:
        """Pauses the execution of the current task and stops the movement of the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotActionException
            If the robot fails to perform the requested action to pause mission execution
            the action to pause will be attempted again until a certain number of retries
        RobotException
            Will catch other RobotExceptions and retry to pause the mission

        """
        raise NotImplementedError

    @abstractmethod
    def resume(self) -> None:
        """Resumes the execution of the current task and continues the rest of the mission.

        Returns
        -------
        None

        Raises
        ------
        RobotActionException
            If the robot fails to perform the requested action to resume mission execution
            the action to resume will be attempted again until a certain number of retries
        RobotException
            Will catch other RobotExceptions and retry to resume the mission

        """
        raise NotImplementedError

    @abstractmethod
    def get_inspection(self, task: InspectionTask) -> Inspection:
        """Return the inspection connected to the given task.

        Parameters
        ----------
        task : InspectionTask

        Returns
        -------
        Sequence[InspectionResult]
            List containing all the inspection results connected to the given task

        Raises
        ------
        RobotRetrieveInspectionException
            If the robot package is unable to retrieve the inspections for the relevant
            mission or task an error message is logged and the state machine continues
        RobotException
            Catches other RobotExceptions that lead to the same result as a
            RobotRetrieveInspectionException

        """
        raise NotImplementedError

    @abstractmethod
    def register_inspection_callback(
        self, callback_function: Callable[[Inspection, Mission], None]
    ) -> None:
        """Register a function which should be run when inspection data is received
        asynchronously. This function should expect to receive an Inspection from.

        Parameters
        ----------
        callback_function : Callable[[Inspection, Mission], None]

        Returns
        -------
        None

        """
        raise NotImplementedError

    @abstractmethod
    def initialize(self, params: InitializeParams) -> None:
        """Initializes the robot. The initialization needed is robot dependent and the
        function can be a simple return statement if no initialization is needed for the
        robot.

        Parameters
        ----------
        params: InitializeParams

        Returns
        -------
        None

        Raises
        ------
        RobotInitializeException
            If the robot package is unable to initialize the robot correctly the mission
            will be cancelled
        RobotException
            Catches other RobotExceptions that might have occurred during initialization
            where the result is that the mission is cancelled

        """
        raise NotImplementedError

    @abstractmethod
    def generate_media_config(self) -> MediaConfig:
        """
        Generate a JSON containing the url and token needed to establish a media stream
        connection to a robot.

        Returns
        -------
        MediaConfig
            An object containing the connection information for a media stream connection
        """
        raise NotImplementedError

    @abstractmethod
    def get_telemetry_publishers(
        self, queue: Queue, isar_id: str, robot_name: str
    ) -> List[Thread]:
        """
        Set up telemetry publisher threads to publish regular updates for pose, battery
        level etc. from the robot to the MQTT broker. The publishers on the robot side
        will use the queue to pass messages to the MQTT Client on the ISAR side.

        The isar_id is passed to the robot to ensure the messages are published to the
        correct topics.

        Note that this functionality will only be utilized if MQTT is enabled in the
        settings.

        Returns
        -------
        List[Thread]
            List containing all threads that will be started to publish telemetry.
        """
        raise NotImplementedError

    @abstractmethod
    def robot_status(self) -> RobotStatus:
        """
        Method which returns an enum indicating if the robot package is able to reach
        the interface which is used to communicate with the robot. This is further used
        by ISAR to indicate whether the ISAR instance is fully functional and may be
        used by other systems.

        Returns
        -------
        RobotStatus
            Enum indicating if the robot may be reached by the isar-robot package.

        Raises
        -------
        RobotCommunicationException
            Raised if the robot package is unable to communicate with the robot API
        RobotAPIException
            Raised if the robot package is able to communicate with the API but an error
            occurred while interpreting the response
        RobotException
            Catches other RobotExceptions that may have occurred while retrieving the
            robot status
            At this point ISAR will attempt to request the robot status again
        """
        raise NotImplementedError



import logging
import os
from queue import Empty, Queue

import backoff
from paho.mqtt import client as mqtt
from paho.mqtt.client import Client

from isar.config.settings import settings
from robot_interface.telemetry.mqtt_client import MqttClientInterface


def _on_success(data: dict) -> None:
    logging.getLogger("mqtt_client").info("Connected to MQTT Broker")
    logging.getLogger("mqtt_client").debug(
        f"Elapsed time: {data['elapsed']}, Tries: {data['tries']}"
    )


def _on_backoff(data: dict) -> None:
    logging.getLogger("mqtt_client").warning(
        f"Failed to connect, retrying in {data['wait']} seconds"
    )


def _on_giveup(data: dict) -> None:
    logging.getLogger("mqtt_client").error(
        "Failed to connect to MQTT Broker within set backoff strategy."
    )


class MqttClient(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.logger = logging.getLogger("mqtt_client")
        self.logger.setLevel("INFO")
        self.mqtt_queue: Queue = mqtt_queue

        username: str = settings.MQTT_USERNAME
        password: str = ""
        try:
            password = os.environ["ISAR_MQTT_PASSWORD"]
        except KeyError:
            self.logger.warning(
                "Failed to retrieve ISAR_MQTT_PASSWORD from environment. Attempting "
                "with empty string as password."
            )

        self.host: str = settings.MQTT_HOST

        # Fix for mqtt running on localhost in docker
        if "IS_DOCKER" in os.environ and self.host == "localhost":
            self.host = "host.docker.internal"

        self.port: int = settings.MQTT_PORT

        self.client: Client = Client(mqtt.CallbackAPIVersion.VERSION1)

        self.client.enable_logger(logger=self.logger)

        dirname = os.path.dirname(__file__)

        if settings.MQTT_SSL_ENABLED:
            cert_path = os.path.join(dirname, "../../../config/certs/ca-cert.pem")
            self.client.tls_set(ca_certs=cert_path)

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect

        self.client.username_pw_set(username=username, password=password)

    def run(self) -> None:
        self.connect(host=self.host, port=self.port)
        self.client.loop_start()

        while True:
            if not self.client.is_connected():
                continue
            try:
                topic, payload, qos, retain = self.mqtt_queue.get(timeout=1)
            except Empty:
                continue

            self.publish(topic=topic, payload=payload, qos=qos, retain=retain)

    def on_connect(self, client, userdata, flags, rc):
        self.logger.info("Connection returned result: " + mqtt.connack_string(rc))

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.logger.warning("Unexpected disconnection from MQTT Broker")

    @backoff.on_exception(
        backoff.expo,
        ConnectionRefusedError,
        max_time=300,
        on_success=_on_success,
        on_backoff=_on_backoff,
        on_giveup=_on_giveup,
    )
    def connect(self, host: str, port: int) -> None:
        self.logger.info("Attempting to connect to MQTT Broker")
        self.logger.info(f"Host: {host}, Port: {port}")
        self.client.connect(host=host, port=port)

    def publish(self, topic: str, payload: str, qos: int = 0, retain: bool = False):
        self.logger.debug(f"Publishing message to topic: {topic}")
        self.client.publish(topic=topic, payload=payload, qos=qos, retain=retain)



import logging
from typing import Union

from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceNotFoundError,
)
from azure.identity import ClientSecretCredential, DefaultAzureCredential
from azure.keyvault.secrets import KeyVaultSecret, SecretClient

from isar.config.keyvault.keyvault_error import KeyvaultError


class Keyvault:
    def __init__(
        self,
        keyvault_name: str,
        client_id: str = None,
        client_secret: str = None,
        tenant_id: str = None,
    ):
        self.name = keyvault_name
        self.url = "https://" + keyvault_name + ".vault.azure.net"
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.logger = logging.getLogger("API")
        self.client: SecretClient = None

    def get_secret(self, secret_name: str) -> KeyVaultSecret:
        secret_client: SecretClient = self.get_secret_client()
        try:
            secret: KeyVaultSecret = secret_client.get_secret(name=secret_name)
        except ResourceNotFoundError:
            self.logger.error(
                "Secret '%s' was not found in keyvault '%s'.",
                secret_name,
                self.name,
                exc_info=True,
            )
            raise KeyvaultError  # type: ignore
        except HttpResponseError:
            self.logger.error(
                "An error occurred while retrieving the secret '%s' from keyvault '%s'.",
                secret_name,
                self.name,
                exc_info=True,
            )
            raise KeyvaultError  # type: ignore

        return secret

    def set_secret(self, secret_name: str, secret_value) -> None:
        secret_client: SecretClient = self.get_secret_client()
        try:
            secret_client.set_secret(name=secret_name, value=secret_value)
        except HttpResponseError:
            self.logger.error(
                "An error occurred while setting secret '%s' in keyvault '%s'.",
                secret_name,
                self.name,
                exc_info=True,
            )
            raise KeyvaultError  # type: ignore

    def get_secret_client(self) -> SecretClient:
        if self.client is None:
            try:
                credential: Union[ClientSecretCredential, DefaultAzureCredential]
                if self.client_id and self.client_secret and self.tenant_id:
                    credential = ClientSecretCredential(
                        tenant_id=self.tenant_id,
                        client_id=self.client_id,
                        client_secret=self.client_secret,
                    )
                else:
                    credential = DefaultAzureCredential()
            except ClientAuthenticationError:
                self.logger.error(
                    "Failed to authenticate to Azure while connecting to KeyVault",
                    exc_info=True,
                )
                raise KeyvaultError

            self.client = SecretClient(vault_url=self.url, credential=credential)
        return self.client



from abc import ABCMeta, abstractmethod
from typing import List

from robot_interface.models.mission.task import TASKS, Task


class TaskSelectorInterface(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.tasks: List[TASKS] = None

    def initialize(self, tasks: List[TASKS]) -> None:
        self.tasks = tasks

    @abstractmethod
    def next_task(self) -> TASKS:
        """
        Returns
        -------
        Task
            Returns the next task from the list of tasks

        Raises
        ------
        TaskSelectorStop
            If all tasks have been selected, and the task selection process is complete.
        """
        pass


class TaskSelectorStop(Exception):
    pass



from .inspection import (
    Audio,
    Image,
    ImageMetadata,
    Inspection,
    InspectionMetadata,
    ThermalImage,
    ThermalImageMetadata,
    ThermalVideo,
    ThermalVideoMetadata,
    Video,
    VideoMetadata,
)



import logging
import time
from typing import TYPE_CHECKING, Callable, Optional

from injector import inject
from transitions import State

from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotInitializeException,
)

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Initialize(State):
    @inject
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="initialize", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine

        self.logger = logging.getLogger("state_machine")
        self.initialize_thread: Optional[ThreadedRequest] = None

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        if self.initialize_thread:
            self.initialize_thread.wait_for_thread()
        self.initialize_thread = None

    def _run(self) -> None:
        transition: Callable
        while True:
            if not self.initialize_thread:
                self.initialize_thread = ThreadedRequest(
                    self.state_machine.robot.initialize
                )
                self.initialize_thread.start_thread(
                    self.state_machine.get_initialize_params(),
                    name="State Machine Initialize Robot",
                )

            try:
                self.initialize_thread.get_output()

            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue

            except (RobotInitializeException, RobotException) as e:
                self.state_machine.current_task.error_message = ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                )
                self.logger.error(
                    f"Failed to initialize robot because: {e.error_description}"
                )
                transition = self.state_machine.initialization_failed  # type: ignore
                break

            transition = self.state_machine.initialization_successful  # type: ignore
            break
        transition()



from dataclasses import dataclass
from datetime import datetime
from typing import List

from alitra import Pose
from transitions import State

from robot_interface.models.mission.status import RobotStatus


@dataclass
class TelemetryPayload:
    isar_id: str
    robot_name: str
    timestamp: datetime


@dataclass
class CloudHealthPayload:
    isar_id: str
    robot_name: str
    timestamp: datetime


@dataclass
class TelemetryPosePayload(TelemetryPayload):
    pose: Pose


@dataclass
class TelemetryBatteryPayload(TelemetryPayload):
    battery_level: float


@dataclass
class TelemetryObstacleStatusPayload(TelemetryPayload):
    obstacle_status: bool


@dataclass
class TelemetryPressurePayload(TelemetryPayload):
    pressure_level: float


@dataclass
class DocumentInfo:
    name: str
    url: str


@dataclass
class VideoStream:
    name: str
    url: str
    type: str


@dataclass
class RobotStatusPayload:
    isar_id: str
    robot_name: str
    robot_status: RobotStatus
    previous_robot_status: RobotStatus
    current_isar_state: State
    current_mission_id: str
    current_task_id: str
    current_step_id: str
    timestamp: datetime


@dataclass
class RobotInfoPayload:
    isar_id: str
    robot_name: str
    robot_model: str
    robot_serial_number: str
    robot_asset: str
    documentation: List[DocumentInfo]
    video_streams: List[VideoStream]
    host: str
    port: int
    capabilities: List[str]
    timestamp: datetime


@dataclass
class RobotHeartbeatPayload:
    isar_id: str
    robot_name: str
    timestamp: datetime



from .idle import Idle
from .initialize import Initialize
from .initiate import Initiate
from .monitor import Monitor
from .off import Off
from .offline import Offline
from .paused import Paused
from .stop import Stop



from typing import List

from injector import Module, multiprovider, provider, singleton

from isar.apis.security.authentication import Authenticator
from isar.storage.storage_interface import StorageInterface
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from tests.mocks.blob_storage import StorageMock
from tests.mocks.mqtt_client import MqttClientMock
from tests.mocks.robot_interface import MockRobot


class MockStorageModule(Module):
    @multiprovider
    @singleton
    def provide_storage(self) -> List[StorageInterface]:
        return [StorageMock()]


class MockMqttModule(Module):
    @provider
    @singleton
    def provide_mqtt(self) -> MqttClientInterface:
        return MqttClientMock()


class MockRobotModule(Module):
    @provider
    @singleton
    def provide_robot(self) -> RobotInterface:
        return MockRobot()


class MockNoAuthenticationModule(Module):
    @provider
    @singleton
    def provide_authenticator(self) -> Authenticator:
        return Authenticator(authentication_enabled=False)


class MockAuthenticationModule(Module):
    @provider
    @singleton
    def provide_authenticator(self) -> Authenticator:
        return Authenticator(authentication_enabled=True)



from queue import Queue

import pytest

from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities


class TestQueueUtilities:
    @pytest.mark.parametrize(
        "message, queue_timeout, expected_message",
        [
            (
                "Test",
                10,
                "Test",
            ),
            (None, 1, None),
        ],
    )
    def test_check_queue_with_queue_size_one(
        self, message, queue_timeout, expected_message
    ):
        test_queue = Queue(maxsize=1)
        if message is not None:
            test_queue.put(message)
            message = QueueUtilities.check_queue(test_queue, queue_timeout)
            assert message == expected_message
        else:
            with pytest.raises(QueueTimeoutError):
                QueueUtilities.check_queue(test_queue, queue_timeout)

    def test_clear_queue(self):
        test_queue = Queue(maxsize=2)
        test_queue.put(1)
        test_queue.put(2)
        QueueUtilities.clear_queue(test_queue)
        assert test_queue.empty()



from dataclasses import field
from datetime import datetime
from queue import Queue
from threading import Thread
from typing import Callable, List, Sequence

from alitra import Frame, Orientation, Pose, Position

from robot_interface.models.robots.media import MediaConnectionType
from robot_interface.models.robots.media import MediaConfig
from robot_interface.models.initialize import InitializeParams
from robot_interface.models.inspection.inspection import (
    Image,
    ImageMetadata,
    Inspection,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import InspectionTask, Task
from robot_interface.robot_interface import RobotInterface


class MockRobot(RobotInterface):
    def __init__(
        self,
        mission_status: MissionStatus = MissionStatus.Successful,
        task_status: TaskStatus = TaskStatus.Successful,
        stop: bool = True,
        pose: Pose = Pose(
            position=Position(x=0, y=0, z=0, frame=Frame("robot")),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
            frame=Frame("robot"),
        ),
        robot_status: RobotStatus = RobotStatus.Available,
    ):
        self.mission_status_return_value: MissionStatus = mission_status
        self.task_status_return_value: TaskStatus = task_status
        self.stop_return_value: bool = stop
        self.robot_pose_return_value: Pose = pose
        self.robot_status_return_value: RobotStatus = robot_status

    def initiate_mission(self, mission: Mission) -> None:
        return

    def initiate_task(self, task: Task) -> None:
        return

    def task_status(self, task_id: str) -> TaskStatus:
        return self.task_status_return_value

    def stop(self) -> None:
        return

    def pause(self) -> None:
        return

    def resume(self) -> None:
        return

    def get_inspection(self, task: InspectionTask) -> Inspection:
        image: Image = Image(mock_image_metadata())
        image.data = b"Some binary image data"
        return image

    def generate_media_config(self) -> MediaConfig:
        return MediaConfig(
            url="mockURL",
            token="mockToken",
            media_connection_type=MediaConnectionType.LiveKit,
        )

    def register_inspection_callback(
        self, callback_function: Callable[[Inspection, Mission], None]
    ) -> None:
        return

    def initialize(self, params: InitializeParams) -> None:
        return

    def get_telemetry_publishers(
        self, queue: Queue, isar_id: str, robot_name: str
    ) -> List[Thread]:
        return []

    def robot_status(self) -> RobotStatus:
        return self.robot_status_return_value


def mock_image_metadata() -> ImageMetadata:
    return ImageMetadata(
        start_time=datetime.now(),
        pose=Pose(
            Position(0, 0, 0, Frame("robot")),
            Orientation(0, 0, 0, 1, Frame("robot")),
            Frame("robot"),
        ),
        file_type="jpg",
    )


class MockRobotIdleToOfflineToIdleTest(MockRobot):
    def __init__(self):
        self.first = True

    def robot_status(self) -> RobotStatus:
        if self.first:
            self.first = False
            return RobotStatus.Offline

        return RobotStatus.Available



from enum import Enum


class MissionStatus(str, Enum):
    NotStarted: str = "not_started"
    InProgress: str = "in_progress"
    Paused: str = "paused"
    Failed: str = "failed"
    Cancelled: str = "cancelled"
    Successful: str = "successful"
    PartiallySuccessful: str = "partially_successful"


class TaskStatus(str, Enum):
    NotStarted: str = "not_started"
    InProgress: str = "in_progress"
    Paused: str = "paused"
    Failed: str = "failed"
    Cancelled: str = "cancelled"
    Successful: str = "successful"
    PartiallySuccessful: str = "partially_successful"


class RobotStatus(Enum):
    Available: str = "available"
    Busy: str = "busy"
    Offline: str = "offline"
    Blocked: str = "blocked"



from typing import Tuple

from robot_interface.robot_interface import RobotInterface


def interface_test(Robot: RobotInterface) -> None:
    for func in dir(RobotInterface):
        if not callable(getattr(Robot, func)) or func.startswith("__"):
            continue

        arg_robot: int = getattr(Robot, func).__code__.co_argcount
        arg_isar: int = getattr(RobotInterface, func).__code__.co_argcount

        args_robot: Tuple[str] = getattr(Robot, func).__code__.co_varnames[:arg_robot]
        args_isar: Tuple[str] = getattr(RobotInterface, func).__code__.co_varnames[
            :arg_isar
        ]

        assert arg_robot == arg_isar
        assert args_robot == args_isar



import logging

from azure.core.exceptions import ClientAuthenticationError
from azure.identity import DefaultAzureCredential


class AzureCredentials:
    @staticmethod
    def get_azure_credentials():
        try:
            return DefaultAzureCredential()
        except ClientAuthenticationError as e:
            logging.error(e.message)
            raise e



import json
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from queue import Empty, Queue
from typing import List, Union

from injector import inject

from isar.config.settings import settings
from isar.models.communication.queues import Queues
from isar.storage.storage_interface import StorageException, StorageInterface
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.utilities.json_service import EnhancedJSONEncoder


@dataclass
class UploaderQueueItem:
    inspection: Inspection
    mission: Mission
    storage_handler: StorageInterface
    _retry_count: int
    _next_retry_time: datetime = datetime.now(timezone.utc)

    def increment_retry(self, max_wait_time: int) -> None:
        self._retry_count += 1
        seconds_until_retry: int = min(2**self._retry_count, max_wait_time)
        self._next_retry_time = datetime.now(timezone.utc) + timedelta(
            seconds=seconds_until_retry
        )

    def get_retry_count(self) -> int:
        return self._retry_count

    def is_ready_for_upload(self) -> bool:
        return datetime.now(timezone.utc) >= self._next_retry_time

    def seconds_until_retry(self) -> int:
        return max(
            0, int((self._next_retry_time - datetime.now(timezone.utc)).total_seconds())
        )


class Uploader:
    @inject
    def __init__(
        self,
        queues: Queues,
        storage_handlers: List[StorageInterface],
        mqtt_publisher: MqttClientInterface,
        max_wait_time: int = settings.UPLOAD_FAILURE_MAX_WAIT,
        max_retry_attempts: int = settings.UPLOAD_FAILURE_ATTEMPTS_LIMIT,
    ) -> None:
        """Initializes the uploader.

        Parameters
        ----------
        queues : Queues
            Queues used for cross-thread communication.
        storage_handlers : List[StorageInterface]
            List of handlers for different upload options
        max_wait_time : float
            The maximum wait time between two retries (exponential backoff)
        max_retry_attempts : int
            Maximum attempts to retry an upload when it fails
        """
        self.upload_queue: Queue = queues.upload_queue
        self.storage_handlers: List[StorageInterface] = storage_handlers
        self.mqtt_publisher = mqtt_publisher

        self.max_wait_time = max_wait_time
        self.max_retry_attempts = max_retry_attempts
        self._internal_upload_queue: List[UploaderQueueItem] = []

        self.logger = logging.getLogger("uploader")

    def run(self) -> None:
        self.logger.info("Started uploader")
        while True:
            inspection: Inspection
            mission: Mission
            try:
                if self._internal_upload_queue:
                    self._process_upload_queue()

                inspection, mission = self.upload_queue.get(timeout=1)

                if not mission:
                    self.logger.warning(
                        "Failed to upload missing mission from upload queue"
                    )
                    continue

                # If new item from thread queue, add one per handler to internal queue:
                for storage_handler in self.storage_handlers:
                    new_item: UploaderQueueItem = UploaderQueueItem(
                        inspection, mission, storage_handler, _retry_count=-1
                    )
                    self._internal_upload_queue.append(new_item)
            except Empty:
                continue

    def _upload(self, upload_item: UploaderQueueItem) -> Union[str, dict]:
        inspection_path: Union[str, dict] = ""
        try:
            inspection_path = upload_item.storage_handler.store(
                inspection=upload_item.inspection, mission=upload_item.mission
            )
            self.logger.info(
                f"Storage handler: {type(upload_item.storage_handler).__name__} "
                f"uploaded inspection {str(upload_item.inspection.id)[:8]}"
            )
            self._internal_upload_queue.remove(upload_item)
        except StorageException:
            if upload_item.get_retry_count() < self.max_retry_attempts:
                upload_item.increment_retry(self.max_wait_time)
                self.logger.warning(
                    f"Storage handler: {type(upload_item.storage_handler).__name__} "
                    f"failed to upload inspection: "
                    f"{str(upload_item.inspection.id)[:8]}. "
                    f"Retrying in {upload_item.seconds_until_retry()}s."
                )
            else:
                self._internal_upload_queue.remove(upload_item)
                self.logger.error(
                    f"Storage handler: {type(upload_item.storage_handler).__name__} "
                    f"exceeded max retries to upload inspection: "
                    f"{str(upload_item.inspection.id)[:8]}. Aborting upload."
                )
        return inspection_path

    def _process_upload_queue(self) -> None:
        ready_items: List[UploaderQueueItem] = [
            item for item in self._internal_upload_queue if item.is_ready_for_upload()
        ]
        for item in ready_items:
            inspection_path = self._upload(item)
            self._publish_inspection_result(
                inspection=item.inspection, inspection_path=inspection_path
            )

    def _publish_inspection_result(
        self, inspection: Inspection, inspection_path: Union[str, dict]
    ) -> None:
        """Publishes the reference of the inspection result to the MQTT Broker
        along with the analysis type
        """
        if not self.mqtt_publisher:
            return
        payload: str = json.dumps(
            {
                "isar_id": settings.ISAR_ID,
                "robot_name": settings.ROBOT_NAME,
                "inspection_id": inspection.id,
                "inspection_path": inspection_path,
                "installation_code": settings.PLANT_SHORT_NAME,
                "analysis_to_be_run": inspection.metadata.analysis_type,
                "timestamp": datetime.now(timezone.utc),
            },
            cls=EnhancedJSONEncoder,
        )
        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_INSPECTION_RESULT,
            payload=payload,
            qos=1,
            retain=True,
        )



import logging
from http import HTTPStatus
from typing import Optional

from alitra import Pose
from fastapi import Body, HTTPException, Path
from injector import inject

from isar.apis.models import InputPose, StartMissionResponse
from isar.apis.models.models import (
    ControlMissionResponse,
    RobotInfoResponse,
    TaskResponse,
)
from isar.apis.models.start_mission_definition import (
    StartMissionDefinition,
    to_isar_mission,
)
from isar.config.settings import robot_settings, settings
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TASKS, Localize, MoveArm, ReturnToHome


class SchedulingController:
    @inject
    def __init__(
        self,
        scheduling_utilities: SchedulingUtilities,
    ):
        self.scheduling_utilities: SchedulingUtilities = scheduling_utilities
        self.logger = logging.getLogger("api")

    def start_mission_by_id(
        self,
        mission_id: str = Path(
            alias="id",
            title="Mission ID",
            description="ID-number for predefined mission",
        ),
        initial_pose: Optional[InputPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial "
            "localization of robot",
            embed=True,
        ),
        return_pose: Optional[InputPose] = Body(
            default=None,
            description="End pose of the mission. The robot return to the specified "
            "pose after finishing all inspections",
            embed=True,
        ),
    ) -> StartMissionResponse:
        self.logger.info(f"Received request to start mission with id {mission_id}")

        state: States = self.scheduling_utilities.get_state()
        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        mission: Mission = self.scheduling_utilities.get_mission(mission_id)
        if return_pose:
            pose: Pose = return_pose.to_alitra_pose()
            mission.tasks.append(ReturnToHome(pose=pose))

        self.scheduling_utilities.verify_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )

        initial_pose_alitra: Optional[Pose] = (
            initial_pose.to_alitra_pose() if initial_pose else None
        )

        self.logger.info(f"Starting mission with ISAR Mission ID: '{mission.id}'")

        self.scheduling_utilities.start_mission(
            mission=mission, initial_pose=initial_pose_alitra
        )

        return self._api_response(mission)

    def start_mission(
        self,
        mission_definition: StartMissionDefinition = Body(
            default=None,
            embed=True,
            title="Mission Definition",
            description="Description of the mission in json format",
        ),
        initial_pose: Optional[InputPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial "
            "localization of robot",
            embed=True,
        ),
        return_pose: Optional[InputPose] = Body(
            default=None,
            description="End pose of the mission. The robot return to the specified "
            "pose after finishing all inspections",
            embed=True,
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new mission")

        if not mission_definition:
            error_message: str = (
                "Unprocessable entity - 'mission_definition' empty or invalid"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY, detail=error_message
            )

        state: States = self.scheduling_utilities.get_state()
        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        try:
            mission: Mission = to_isar_mission(mission_definition)
        except MissionPlannerError as e:
            error_message = f"Bad Request - Cannot create ISAR mission: {e}"
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        self.scheduling_utilities.verify_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )
        if return_pose:
            pose: Pose = return_pose.to_alitra_pose()
            mission.tasks.append(ReturnToHome(pose=pose))

        initial_pose_alitra: Optional[Pose] = (
            initial_pose.to_alitra_pose() if initial_pose else None
        )

        self.logger.info(f"Starting mission: {mission.id}")
        self.scheduling_utilities.start_mission(
            mission=mission, initial_pose=initial_pose_alitra
        )
        return self._api_response(mission)

    def pause_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to pause current mission")

        state: States = self.scheduling_utilities.get_state()

        if state not in [
            States.Monitor,
            States.Initiate,
        ]:
            error_message = (
                f"Conflict - Pause command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=error_message,
            )

        pause_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.pause_mission()
        )
        return pause_mission_response

    def resume_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to resume current mission")

        state: States = self.scheduling_utilities.get_state()

        if state != States.Paused:
            error_message = (
                f"Conflict - Resume command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        resume_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.resume_mission()
        )
        return resume_mission_response

    def stop_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to stop current mission")

        state: States = self.scheduling_utilities.get_state()

        if state == States.Off:
            error_message = (
                f"Conflict - Stop command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        stop_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.stop_mission()
        )
        return stop_mission_response

    def drive_to(
        self,
        target_pose: InputPose = Body(
            default=None,
            title="Target Pose",
            description="The target pose for the drive_to task",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new drive-to mission")

        state: States = self.scheduling_utilities.get_state()

        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        pose: Pose = target_pose.to_alitra_pose()
        mission: Mission = Mission(tasks=[ReturnToHome(pose=pose)])

        self.logger.info(
            f"Starting drive to mission with ISAR Mission ID: '{mission.id}'"
        )
        self.scheduling_utilities.start_mission(mission=mission, initial_pose=None)
        return self._api_response(mission)

    def start_localization_mission(
        self,
        localization_pose: InputPose = Body(
            default=None,
            embed=True,
            title="Localization Pose",
            description="The current position of the robot",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new localization mission")

        state: States = self.scheduling_utilities.get_state()

        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        pose: Pose = localization_pose.to_alitra_pose()
        mission: Mission = Mission(tasks=[Localize(localization_pose=pose)])

        self.logger.info(
            f"Starting localization mission with ISAR Mission ID: '{mission.id}'"
        )
        self.scheduling_utilities.start_mission(
            mission=mission,
            initial_pose=None,
        )
        return self._api_response(mission)

    def start_move_arm_mission(
        self,
        arm_pose_literal: str = Path(
            ...,
            alias="arm_pose_literal",
            title="Arm pose literal",
            description="Arm pose as a literal",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new move arm mission")

        if not robot_settings.VALID_ARM_POSES:
            error_message: str = (
                f"Received a request to move the arm but the robot "
                f"{settings.ROBOT_NAME} does not support moving an arm"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST, detail=error_message
            )

        if arm_pose_literal not in robot_settings.VALID_ARM_POSES:
            error_message = (
                f"Received a request to move the arm but the arm pose "
                f"{arm_pose_literal} is not supported by the robot "
                f"{settings.ROBOT_NAME}"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST, detail=error_message
            )

        state: States = self.scheduling_utilities.get_state()

        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        mission: Mission = Mission(tasks=[MoveArm(arm_pose=arm_pose_literal)])

        self.logger.info(
            f"Starting move arm mission with ISAR Mission ID: '{mission.id}'"
        )
        self.scheduling_utilities.start_mission(
            mission=mission,
            initial_pose=None,
        )
        return self._api_response(mission)

    def _api_response(self, mission: Mission) -> StartMissionResponse:
        return StartMissionResponse(
            id=mission.id,
            tasks=[self._task_api_response(task) for task in mission.tasks],
        )

    def _task_api_response(self, task: TASKS) -> TaskResponse:
        return TaskResponse(id=task.id, tag_id=task.tag_id, type=task.type)



from alitra import Frame, Position

from robot_interface.models.mission.task import ReturnToHome, TakeImage
from tests.mocks.pose import MockPose


class MockTask:
    @staticmethod
    def return_home() -> ReturnToHome:
        return ReturnToHome(pose=MockPose.default_pose())

    @staticmethod
    def take_image() -> TakeImage:
        target_pose = Position(x=1, y=1, z=1, frame=Frame("robot"))
        robot_pose = Position(x=0, y=0, z=1, frame=Frame("robot"))
        return TakeImage(target=target_pose, robot_pose=robot_pose)



import logging
import time
from typing import TYPE_CHECKING, Optional

from transitions import State

from isar.config.settings import settings
from isar.models.communication.message import StartMissionMessage
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions.robot_exceptions import RobotException
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Idle(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="idle", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.robot_status_thread: Optional[ThreadedRequest] = None
        self.last_robot_status_poll_time: float = time.time()

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        if self.robot_status_thread:
            self.robot_status_thread.wait_for_thread()
        self.robot_status_thread = None

    def _is_ready_to_poll_for_status(self) -> bool:
        time_since_last_robot_status_poll = (
            time.time() - self.last_robot_status_poll_time
        )
        return (
            time_since_last_robot_status_poll > settings.ROBOT_API_STATUS_POLL_INTERVAL
        )

    def _run(self) -> None:
        while True:
            if self.state_machine.should_stop_mission():
                transition = self.state_machine.stop  # type: ignore
                break
            start_mission: Optional[StartMissionMessage] = (
                self.state_machine.should_start_mission()
            )
            if start_mission:
                self.state_machine.start_mission(
                    mission=start_mission.mission,
                    initial_pose=start_mission.initial_pose,
                )
                transition = self.state_machine.mission_started  # type: ignore
                break
            time.sleep(self.state_machine.sleep_time)

            if not self._is_ready_to_poll_for_status():
                continue

            if not self.robot_status_thread:
                self.robot_status_thread = ThreadedRequest(
                    request_func=self.state_machine.robot.robot_status
                )
                self.robot_status_thread.start_thread(
                    name="State Machine Offline Get Robot Status"
                )

            try:
                robot_status: RobotStatus = self.robot_status_thread.get_output()
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue

            except RobotException as e:
                self.logger.error(
                    f"Failed to get robot status because: {e.error_description}"
                )

            self.last_robot_status_poll_time = time.time()

            if robot_status == RobotStatus.Offline:
                transition = self.state_machine.robot_turned_offline  # type: ignore
                break

            self.robot_status_thread = None

        transition()



from enum import Enum


# Did you write your own isar-robot package and would like to have it included here?
# Open a pull request to the ISAR repository!
class RobotModel(Enum):
    TaurobInspector: str = "TaurobInspector"
    TaurobOperator: str = "TaurobOperator"
    ExR2: str = "ExR2"
    Robot: str = "Robot"  # This corresponds to the mock in isar_robot
    Turtlebot: str = "Turtlebot"
    AnymalX: str = "AnymalX"
    AnymalD: str = "AnymalD"



import json
import logging
import queue
from collections import deque
from datetime import datetime, timezone
from typing import Deque, List, Optional

from alitra import Pose
from injector import inject
from transitions import Machine
from transitions.core import State

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.mission_planner.task_selector_interface import (
    TaskSelectorInterface,
    TaskSelectorStop,
)
from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues.queues import Queues
from isar.state_machine.states import (
    Idle,
    Initialize,
    Initiate,
    Monitor,
    Off,
    Offline,
    Paused,
    Stop,
)
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.initialize.initialize_params import InitializeParams
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import TASKS
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class StateMachine(object):
    """Handles state transitions for supervisory robot control."""

    @inject
    def __init__(
        self,
        queues: Queues,
        robot: RobotInterface,
        mqtt_publisher: MqttClientInterface,
        task_selector: TaskSelectorInterface,
        sleep_time: float = settings.FSM_SLEEP_TIME,
        run_mission_by_task: bool = settings.RUN_MISSION_BY_TASK,
        stop_robot_attempts_limit: int = settings.STOP_ROBOT_ATTEMPTS_LIMIT,
        transitions_log_length: int = settings.STATE_TRANSITIONS_LOG_LENGTH,
    ):
        """Initializes the state machine.

        Parameters
        ----------
        queues : Queues
            Queues used for API communication.
        robot : RobotInterface
            Instance of robot interface.
        mqtt_publisher : MqttClientInterface
            Instance of MQTT client interface which has a publish function
        sleep_time : float
            Time to sleep in between state machine iterations.
        stop_robot_attempts_limit : int
            Maximum attempts to stop the robot when stop command is received
        transitions_log_length : int
            Length of state transition log list.

        """
        self.logger = logging.getLogger("state_machine")

        self.queues: Queues = queues
        self.robot: RobotInterface = robot
        self.mqtt_publisher: Optional[MqttClientInterface] = mqtt_publisher
        self.task_selector: TaskSelectorInterface = task_selector
        self.run_mission_by_task: bool = run_mission_by_task

        # List of states
        self.stop_state: State = Stop(self)
        self.paused_state: State = Paused(self)
        self.idle_state: State = Idle(self)
        self.initialize_state: State = Initialize(self)
        self.monitor_state: State = Monitor(self)
        self.initiate_state: State = Initiate(self)
        self.off_state: State = Off(self)
        self.offline_state: State = Offline(self)

        self.states: List[State] = [
            self.off_state,
            self.idle_state,
            self.initialize_state,
            self.initiate_state,
            self.monitor_state,
            self.stop_state,
            self.paused_state,
            self.offline_state,
        ]

        self.machine = Machine(self, states=self.states, initial="off", queued=True)
        self.machine.add_transitions(
            [
                {
                    "trigger": "start_machine",
                    "source": self.off_state,
                    "dest": self.idle_state,
                    "before": self._off,
                },
                {
                    "trigger": "initiated",
                    "source": self.initiate_state,
                    "dest": self.monitor_state,
                    "before": self._initiated,
                },
                {
                    "trigger": "pause_full_mission",
                    "source": [self.initiate_state, self.monitor_state],
                    "dest": self.paused_state,
                    "before": self._mission_paused,
                },
                {
                    "trigger": "pause",
                    "source": [self.initiate_state, self.monitor_state],
                    "dest": self.stop_state,
                    "before": self._pause,
                },
                {
                    "trigger": "stop",
                    "source": [
                        self.initiate_state,
                        self.monitor_state,
                        self.idle_state,
                    ],
                    "dest": self.stop_state,
                    "before": self._stop,
                },
                {
                    "trigger": "mission_finished",
                    "source": [
                        self.initiate_state,
                    ],
                    "dest": self.idle_state,
                    "before": self._mission_finished,
                },
                {
                    "trigger": "mission_started",
                    "source": self.idle_state,
                    "dest": self.initialize_state,
                    "before": self._mission_started,
                },
                {
                    "trigger": "initialization_successful",
                    "source": self.initialize_state,
                    "dest": self.initiate_state,
                    "before": self._initialization_successful,
                },
                {
                    "trigger": "initialization_failed",
                    "source": self.initialize_state,
                    "dest": self.idle_state,
                    "before": self._initialization_failed,
                },
                {
                    "trigger": "resume",
                    "source": self.paused_state,
                    "dest": self.initiate_state,
                    "before": self._resume,
                },
                {
                    "trigger": "resume_full_mission",
                    "source": self.paused_state,
                    "dest": self.monitor_state,
                    "before": self._resume,
                },
                {
                    "trigger": "task_finished",
                    "source": self.monitor_state,
                    "dest": self.initiate_state,
                    "before": self._task_finished,
                },
                {
                    "trigger": "full_mission_finished",
                    "source": self.monitor_state,
                    "dest": self.initiate_state,
                    "before": self._full_mission_finished,
                },
                {
                    "trigger": "mission_paused",
                    "source": self.stop_state,
                    "dest": self.paused_state,
                    "before": self._mission_paused,
                },
                {
                    "trigger": "initiate_infeasible",
                    "source": self.initiate_state,
                    "dest": self.initiate_state,
                    "before": self._initiate_infeasible,
                },
                {
                    "trigger": "initiate_failed",
                    "source": self.initiate_state,
                    "dest": self.idle_state,
                    "before": self._initiate_failed,
                },
                {
                    "trigger": "mission_stopped",
                    "source": [self.stop_state, self.paused_state],
                    "dest": self.idle_state,
                    "before": self._mission_stopped,
                },
                {
                    "trigger": "robot_turned_offline",
                    "source": [self.idle_state],
                    "dest": self.offline_state,
                    "before": self._offline,
                },
                {
                    "trigger": "robot_turned_online",
                    "source": self.offline_state,
                    "dest": self.idle_state,
                    "before": self._online,
                },
            ]
        )

        self.stop_robot_attempts_limit: int = stop_robot_attempts_limit
        self.sleep_time: float = sleep_time

        self.stopped: bool = False
        self.current_mission: Optional[Mission] = None
        self.current_task: Optional[TASKS] = None
        self.initial_pose: Optional[Pose] = None

        self.current_state: State = States(self.state)  # type: ignore

        self.transitions_log_length: int = transitions_log_length
        self.transitions_list: Deque[States] = deque([], self.transitions_log_length)

    #################################################################################
    # Transition Callbacks
    def _initialization_successful(self) -> None:
        return

    def _initialization_failed(self) -> None:
        self.queues.start_mission.output.put(False)
        self._finalize()

    def _initiated(self) -> None:
        if self.run_mission_by_task:
            self.current_task.status = TaskStatus.InProgress
        self.current_mission.status = MissionStatus.InProgress
        self.publish_task_status(task=self.current_task)
        self.logger.info(
            f"Successfully initiated "
            f"{type(self.current_task).__name__} "
            f"task: {str(self.current_task.id)[:8]}"
        )

    def _pause(self) -> None:
        return

    def _off(self) -> None:
        return

    def _offline(self) -> None:
        return

    def _online(self) -> None:
        return

    def _resume(self) -> None:
        self.logger.info(f"Resuming mission: {self.current_mission.id}")
        self.current_mission.status = MissionStatus.InProgress
        self.current_mission.error_message = None
        self.current_task.status = TaskStatus.InProgress

        self.publish_mission_status()
        self.publish_task_status(task=self.current_task)

        resume_mission_response: ControlMissionResponse = (
            self._make_control_mission_response()
        )
        self.queues.resume_mission.output.put(resume_mission_response)

        self.robot.resume()

    def _mission_finished(self) -> None:
        fail_statuses: List[TaskStatus] = [
            TaskStatus.Cancelled,
            TaskStatus.Failed,
        ]
        partially_fail_statuses = fail_statuses + [TaskStatus.PartiallySuccessful]

        if len(self.current_mission.tasks) == 0:
            self.current_mission.status = MissionStatus.Successful
        elif all(task.status in fail_statuses for task in self.current_mission.tasks):
            self.current_mission.error_message = ErrorMessage(
                error_reason=None,
                error_description="The mission failed because all tasks in the mission "
                "failed",
            )
            self.current_mission.status = MissionStatus.Failed
        elif any(
            task.status in partially_fail_statuses
            for task in self.current_mission.tasks
        ):
            self.current_mission.status = MissionStatus.PartiallySuccessful
        else:
            self.current_mission.status = MissionStatus.Successful
        self._finalize()

    def _mission_started(self) -> None:
        self.queues.start_mission.output.put(True)
        self.logger.info(
            f"Initialization successful. Starting new mission: "
            f"{self.current_mission.id}"
        )
        self.log_mission_overview(mission=self.current_mission)

        self.current_mission.status = MissionStatus.InProgress
        self.publish_mission_status()
        self.current_task = self.task_selector.next_task()
        if self.current_task is None:
            self._mission_finished()
        else:
            self.current_task.status = TaskStatus.InProgress
            self.publish_task_status(task=self.current_task)

    def _task_finished(self) -> None:
        self.publish_task_status(task=self.current_task)
        self.current_task.update_task_status()
        self.iterate_current_task()

    def _full_mission_finished(self) -> None:
        self.current_task = None

    def _mission_paused(self) -> None:
        self.logger.info(f"Pausing mission: {self.current_mission.id}")
        self.current_mission.status = MissionStatus.Paused
        self.current_task.status = TaskStatus.Paused

        paused_mission_response: ControlMissionResponse = (
            self._make_control_mission_response()
        )
        self.queues.pause_mission.output.put(paused_mission_response)

        self.publish_mission_status()
        self.publish_task_status(task=self.current_task)

        self.robot.pause()

    def _stop(self) -> None:
        self.stopped = True

    def _initiate_failed(self) -> None:
        self.current_task.status = TaskStatus.Failed
        self.current_mission.status = MissionStatus.Failed
        self.publish_task_status(task=self.current_task)
        self._finalize()

    def _initiate_infeasible(self) -> None:
        if self.run_mission_by_task:
            self.current_task.status = TaskStatus.Failed
            self.publish_task_status(task=self.current_task)
            self.iterate_current_task()

    def _mission_stopped(self) -> None:
        if self.current_mission is None:
            self._queue_empty_response()
            self.reset_state_machine()
            return

        self.current_mission.status = MissionStatus.Cancelled

        for task in self.current_mission.tasks:
            if task.status in [
                TaskStatus.NotStarted,
                TaskStatus.InProgress,
                TaskStatus.Paused,
            ]:
                task.status = TaskStatus.Cancelled

        stopped_mission_response: ControlMissionResponse = (
            self._make_control_mission_response()
        )
        self.queues.stop_mission.output.put(stopped_mission_response)

        self.publish_task_status(task=self.current_task)
        self._finalize()

    #################################################################################

    def _finalize(self) -> None:
        self.publish_mission_status()
        self.log_mission_overview(mission=self.current_mission)
        state_transitions: str = ", ".join(
            [
                f"\n  {transition}" if (i + 1) % 10 == 0 else f"{transition}"
                for i, transition in enumerate(list(self.transitions_list))
            ]
        )
        self.logger.info(f"State transitions:\n  {state_transitions}")
        self.reset_state_machine()

    def begin(self):
        """Starts the state machine.

        Transitions into idle state.

        """
        self.to_idle()

    def iterate_current_task(self):
        if self.current_task.is_finished():
            try:
                self.current_task = self.task_selector.next_task()
                self.current_task.status = TaskStatus.InProgress
                self.publish_task_status(task=self.current_task)
            except TaskSelectorStop:
                # Indicates that all tasks are finished
                self.current_task = None

    def update_state(self):
        """Updates the current state of the state machine."""
        self.current_state = States(self.state)
        self.send_state_status()
        self._log_state_transition(self.current_state)
        self.logger.info(f"State: {self.current_state}")
        self.publish_status()

    def reset_state_machine(self) -> None:
        self.logger.info("Resetting state machine")
        self.stopped = False
        self.current_task = None
        self.current_mission = None
        self.initial_pose = None

    def start_mission(self, mission: Mission, initial_pose: Pose):
        """Starts a scheduled mission."""
        self.current_mission = mission
        self.initial_pose = initial_pose

        self.task_selector.initialize(tasks=self.current_mission.tasks)

    def get_initialize_params(self):
        return InitializeParams(initial_pose=self.initial_pose)

    def should_start_mission(self) -> Optional[StartMissionMessage]:
        try:
            return self.queues.start_mission.input.get(block=False)
        except queue.Empty:
            return None

    def should_stop_mission(self) -> bool:
        try:
            return self.queues.stop_mission.input.get(block=False)
        except queue.Empty:
            return False

    def should_pause_mission(self) -> bool:
        try:
            return self.queues.pause_mission.input.get(block=False)
        except queue.Empty:
            return False

    def should_resume_mission(self) -> bool:
        try:
            return self.queues.resume_mission.input.get(block=False)
        except queue.Empty:
            return False

    def send_state_status(self):
        self.queues.state.update(self.current_state)

    def publish_mission_status(self) -> None:
        if not self.mqtt_publisher:
            return

        error_message: Optional[ErrorMessage] = None
        if self.current_mission:
            if self.current_mission.error_message:
                error_message = self.current_mission.error_message
        payload: str = json.dumps(
            {
                "isar_id": settings.ISAR_ID,
                "robot_name": settings.ROBOT_NAME,
                "mission_id": self.current_mission.id if self.current_mission else None,
                "status": self.current_mission.status if self.current_mission else None,
                "error_reason": error_message.error_reason if error_message else None,
                "error_description": (
                    error_message.error_description if error_message else None
                ),
                "timestamp": datetime.now(timezone.utc),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_MISSION,
            payload=payload,
            qos=1,
            retain=True,
        )

    def publish_task_status(self, task: TASKS) -> None:
        """Publishes the task status to the MQTT Broker"""
        if not self.mqtt_publisher:
            return

        error_message: Optional[ErrorMessage] = None
        if task:
            if task.error_message:
                error_message = task.error_message

        payload: str = json.dumps(
            {
                "isar_id": settings.ISAR_ID,
                "robot_name": settings.ROBOT_NAME,
                "mission_id": self.current_mission.id if self.current_mission else None,
                "task_id": task.id if task else None,
                "status": task.status if task else None,
                "task_type": task.type,
                "error_reason": error_message.error_reason if error_message else None,
                "error_description": (
                    error_message.error_description if error_message else None
                ),
                "timestamp": datetime.now(timezone.utc),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_TASK,
            payload=payload,
            qos=1,
            retain=True,
        )

    def publish_status(self) -> None:
        if not self.mqtt_publisher:
            return
        payload: str = json.dumps(
            {
                "isar_id": settings.ISAR_ID,
                "robot_name": settings.ROBOT_NAME,
                "status": self._current_status(),
                "timestamp": datetime.now(timezone.utc),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_STATUS,
            payload=payload,
            qos=1,
            retain=True,
        )

    def _current_status(self) -> RobotStatus:
        if self.current_state == States.Idle:
            return RobotStatus.Available
        elif self.current_state == States.Offline:
            return RobotStatus.Offline
        else:
            return RobotStatus.Busy

    def _log_state_transition(self, next_state):
        """Logs all state transitions that are not self-transitions."""
        self.transitions_list.append(next_state)

    def log_mission_overview(self, mission: Mission):
        """Log an overview of the tasks in a mission"""
        log_statements: List[str] = []
        for task in mission.tasks:
            log_statements.append(
                f"{type(task).__name__:<20} {str(task.id)[:8]:<32} -- {task.status}"
            )
        log_statement: str = "\n".join(log_statements)

        self.logger.info(f"Mission overview:\n{log_statement}")

    def _make_control_mission_response(self) -> ControlMissionResponse:
        return ControlMissionResponse(
            mission_id=self.current_mission.id,
            mission_status=self.current_mission.status,
            task_id=self.current_task.id,
            task_status=self.current_task.status,
        )

    def _queue_empty_response(self):
        self.queues.stop_mission.output.put(
            ControlMissionResponse(
                mission_id="None",
                mission_status="None",
                task_id="None",
                task_status="None",
            )
        )


def main(state_machine: StateMachine):
    """Starts a state machine instance."""
    state_machine.begin()



import json
import os
from typing import List

import pytest
from alitra import Frame, Orientation, Pose, Position

from isar.apis.models.start_mission_definition import (
    StartMissionDefinition,
    get_duplicate_ids,
    to_isar_mission,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TASKS, Task

task_1: Task = Task(tag_id=None, id="123")
task_2: Task = Task(tag_id=None, id="123")
task_3: Task = Task(tag_id=None, id="123456")
task_4: Task = Task()
task_5: Task = Task()


@pytest.mark.parametrize(
    "item_list, expected_boolean",
    [
        (
            [task_1, task_2, task_3],
            True,
        ),
        (
            [task_1, task_3, task_4, task_5],
            False,
        ),
    ],
)
def test_duplicate_id_check(item_list: List[TASKS], expected_boolean: bool):
    duplicates: List[str] = get_duplicate_ids(item_list)
    has_duplicates: bool = len(duplicates) > 0
    assert has_duplicates == expected_boolean


def test_mission_definition_to_isar_mission():
    dirname = os.path.dirname(__file__)
    filepath = os.path.join(dirname, "example_mission_definition.json")

    with open(filepath) as f:
        datax = json.load(f)
        mission_definition = StartMissionDefinition(**datax)

    generated_mission: Mission = to_isar_mission(mission_definition)
    assert generated_mission.id == "generated_mission_id"
    assert generated_mission.name == "my-mission"
    assert len(generated_mission.tasks) == 1

    task = generated_mission.tasks[0]
    assert task.id == "generated_task_id"
    assert task.robot_pose == Pose(
        position=Position(0.0, 0.0, 0.0, frame=Frame("robot")),
        orientation=Orientation(0.0, 0.0, 0.0, 0.0, frame=Frame("robot")),
        frame=Frame("robot"),
    )
    assert task.type == "take_image"
    assert task.target == Position(0.0, 0.0, 0.0, frame=Frame("robot"))
